#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkevents.h>
#include <X11/Xlib.h>

#define GDK_PIXMAP_SCREEN(pix)	      (GDK_DRAWABLE_IMPL_X11 (((GdkPixmapObject *)pix)->impl)->screen)
#define GDK_PIXMAP_DISPLAY(pix)       (GDK_SCREEN_X11 (GDK_PIXMAP_SCREEN (pix))->display)
#define GDK_PIXMAP_XROOTWIN(pix)      (GDK_SCREEN_X11 (GDK_PIXMAP_SCREEN (pix))->xroot_window)
#define GDK_DRAWABLE_DISPLAY(win)     (GDK_IS_WINDOW (win) ? GDK_WINDOW_DISPLAY (win) : GDK_PIXMAP_DISPLAY (win))
#define GDK_DRAWABLE_SCREEN(win)      (GDK_IS_WINDOW (win) ? GDK_WINDOW_SCREEN (win) : GDK_PIXMAP_SCREEN (win))
#define GDK_DRAWABLE_XROOTWIN(win)    (GDK_IS_WINDOW (win) ? GDK_WINDOW_XROOTWIN (win) : GDK_PIXMAP_XROOTWIN (win))
#define GDK_SCREEN_DISPLAY(screen)    (GDK_SCREEN_X11 (screen)->display)
#define GDK_SCREEN_XROOTWIN(screen)   (GDK_SCREEN_X11 (screen)->xroot_window)
#define GDK_WINDOW_SCREEN(win)	      (GDK_DRAWABLE_IMPL_X11 (((GdkWindowObject *)win)->impl)->screen)
#define GDK_WINDOW_DISPLAY(win)       (GDK_SCREEN_X11 (GDK_WINDOW_SCREEN (win))->display)
#define GDK_WINDOW_XROOTWIN(win)      (GDK_SCREEN_X11 (GDK_WINDOW_SCREEN (win))->xroot_window)
#define GDK_GC_DISPLAY(gc)            (GDK_SCREEN_DISPLAY (GDK_GC_X11(gc)->screen))


#include "jackvst.h"

#include <lash/lash.h>
static	GtkWidget* window;
static	GtkWidget* socket;
static	GtkWidget* vpacker;
static	GtkWidget* hpacker;
static	GtkWidget* bypass_button;
static	GtkWidget* remove_button;
static	GtkWidget* mute_button;
static  GtkWidget* event_box;

gboolean quit = FALSE;

extern lash_client_t * lash_client;

static void
bypass_handler (GtkToggleButton *but, gboolean ptr)
{
	JackVST* jvst = (JackVST*) ptr;
	
	jvst->bypassed = gtk_toggle_button_get_active (but);
}

static void
mute_handler (GtkToggleButton *but, gboolean ptr)
{
	JackVST* jvst = (JackVST*) ptr;
	jvst->muted = gtk_toggle_button_get_active (but);
}

static void
remove_handler (GtkToggleButton *but, gboolean ptr)
{
	JackVST* jvst = (JackVST*) ptr;
	
	jack_deactivate (jvst->client);
	fst_destroy_editor (jvst->fst);
}

static gboolean
configure_handler (GtkWidget* widget, GdkEventConfigure* ev, GtkSocket *socket)
{
	XEvent event;
	gint x, y;

	g_return_if_fail (socket->plug_window != NULL);
	event.xconfigure.type = ConfigureNotify;

	event.xconfigure.event = GDK_WINDOW_XWINDOW (socket->plug_window);
	event.xconfigure.window = GDK_WINDOW_XWINDOW (socket->plug_window);

	/* The ICCCM says that synthetic events should have root relative
	 * coordinates. We still aren't really ICCCM compliant, since
	 * we don't send events when the real toplevel is moved.
	 */
	gdk_error_trap_push ();
	gdk_window_get_origin (socket->plug_window, &x, &y);
	gdk_error_trap_pop ();

	event.xconfigure.x = x;
	event.xconfigure.y = y;
	event.xconfigure.width = GTK_WIDGET(socket)->allocation.width;
	event.xconfigure.height = GTK_WIDGET(socket)->allocation.height;

	event.xconfigure.border_width = 0;
	event.xconfigure.above = None;
	event.xconfigure.override_redirect = False;

	gdk_error_trap_push ();
	XSendEvent (GDK_WINDOW_XDISPLAY (socket->plug_window),
		    GDK_WINDOW_XWINDOW (socket->plug_window),
		    False, StructureNotifyMask, &event);
	gdk_display_sync (gtk_widget_get_display (GTK_WIDGET (socket)));
	gdk_error_trap_pop ();

	return FALSE;
}

void
forward_key_event (GdkEventKey* ev, GtkSocket* socket, JackVST* jvst)
{
	XKeyEvent event;
	Status status;

	g_return_if_fail (socket->plug_window != NULL);
	
	event.type = (ev->type == GDK_KEY_PRESS ? KeyPress : KeyRelease);
	event.display = GDK_WINDOW_XDISPLAY (socket->plug_window);
	event.window = fst_get_XID (jvst->fst);
	event.time = ev->time;
	event.x = 1;
	event.y = 1;
	event.x_root = 1;
	event.y_root = 1;
	event.state = ev->state;
	event.keycode = ev->hardware_keycode;
	event.same_screen = True;
	
	gdk_error_trap_push ();
	XSendEvent (event.display, event.window, False, 0, (XEvent*) &event);
	gdk_display_sync (gtk_widget_get_display (GTK_WIDGET (socket)));
	gdk_error_trap_pop ();
}

static gboolean
destroy_handler (GtkWidget* widget, GdkEventAny* ev, gpointer ptr)
{
	JackVST* jvst = (JackVST*) ptr;
	fst_destroy_editor (jvst->fst);
	//exit (0);
	gtk_main_quit();
	
	return FALSE;
}

int
focus_handler (GtkWidget* widget, GdkEventFocus* ev, gpointer ptr)
{
	if (ev->in) {
		fst_error ("Socket focus in");
	} else {
		fst_error ("Socket focus out");
	}
		       
	return FALSE;
}

////////////////
// LASH
//


void
save_data( JackVST *jvst )
{
	int i;
	lash_config_t *config;

	for( i=0; i<jvst->fst->plugin->numParams; i++ ) {
	    char buf[10];
	    
	    snprintf( buf, 9, "%d", i );

	    config = lash_config_new_with_key( buf );
	    lash_config_set_value_double(config, jvst->fst->plugin->getParameter( jvst->fst->plugin, i ) );
	    lash_send_config(lash_client, config);
	}

}

void
restore_data(lash_config_t * config, JackVST *jvst )
{
	const char *key;

	key = lash_config_get_key(config);

	if (strcmp(key, "dummy") == 0) {
		lash_config_get_value_double(config);
		return;
	}

	jvst->fst->plugin->setParameter( jvst->fst->plugin, atoi( key ), lash_config_get_value_double( config ) );

}
static gboolean
idle_cb(JackVST *jvst)
{
	if (quit) {
		fst_destroy_editor( jvst->fst);
		gtk_main_quit();
		return FALSE;
	}

	if (lash_enabled(lash_client)) {
	    lash_event_t *event;
	    lash_config_t *config;

	    while ((event = lash_get_event(lash_client))) {
		switch (lash_event_get_type(event)) {
		    case LASH_Quit:
			quit = 1;
			lash_event_destroy(event);
			break;
		    case LASH_Restore_Data_Set:
			printf( "lash_restore... \n" );
			lash_send_event(lash_client, event);
			break;
		    case LASH_Save_Data_Set:
			printf( "lash_save... \n" );
			save_data( jvst );
			lash_send_event(lash_client, event);
			break;
		    case LASH_Server_Lost:
			return 1;
		    default:
			printf("%s: receieved unknown LASH event of type %d",
				__FUNCTION__, lash_event_get_type(event));
			lash_event_destroy(event);
			break;
		}
	    }

	    while ((config = lash_get_config(lash_client))) {
		restore_data(config, jvst);
		lash_config_destroy(config);
	    }

	}

	return TRUE;
}

int
manage_vst_plugin (JackVST* jvst)
{
	// create a GtkWindow containing a GtkSocket...
	//
	// notice the order of the functions.
	// you can only add an id to an anchored widget.
	
	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW(window), jvst->handle->name);

	vpacker = gtk_vbox_new (FALSE, 7);
	hpacker = gtk_hbox_new (FALSE, 7);
	bypass_button = gtk_toggle_button_new_with_label ("bypass");
	mute_button = gtk_toggle_button_new_with_label ("mute");
	remove_button = gtk_toggle_button_new_with_label ("remove");

	gtk_signal_connect (GTK_OBJECT(bypass_button), "toggled",
			    GTK_SIGNAL_FUNC(bypass_handler),
			    jvst);

	gtk_signal_connect (GTK_OBJECT(mute_button), "toggled",
			    GTK_SIGNAL_FUNC(mute_handler),
			    jvst);

	gtk_signal_connect (GTK_OBJECT(remove_button), "toggled",
			    GTK_SIGNAL_FUNC(remove_handler),
			    jvst);

	gtk_container_set_border_width (GTK_CONTAINER(hpacker), 3);

	gtk_signal_connect (GTK_OBJECT(window), "delete_event",
			    GTK_SIGNAL_FUNC(destroy_handler),
			    jvst);

	socket = gtk_socket_new ();
	GTK_WIDGET_SET_FLAGS(socket, GTK_CAN_FOCUS);
	
	gtk_box_pack_end   (GTK_BOX(hpacker), bypass_button, FALSE, FALSE, 0);
	gtk_box_pack_end   (GTK_BOX(hpacker), mute_button, FALSE, FALSE, 0);
	// gtk_box_pack_end   (GTK_BOX(hpacker), remove_button, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(vpacker), hpacker, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(vpacker), socket, TRUE, TRUE, 0);

	gtk_container_add (GTK_CONTAINER (window), vpacker);

	// normally every socket should register it self like this.
	gtk_signal_connect (GTK_OBJECT(window), "configure_event",
			    GTK_SIGNAL_FUNC(configure_handler),
			    socket);

	//gtk_socket_add_id (GTK_SOCKET (socket), fst_get_XID (jvst->fst));
	
	fst_show_editor (jvst->fst);

	// but you can show() a GtkSocket only with an id set.
	
 	//gtk_widget_show_all (window);
	//gtk_window_set_focus (GTK_WINDOW(window), socket);


	g_timeout_add(500, (GSourceFunc) idle_cb, jvst);

	gtk_main ();

	return 0;
}

void
gui_init (int *argc, char **argv[])
{
	gtk_init (argc, argv);
}
